@@ -1,14 +1,23 @@
-from io import BytesIO
+from io import BufferedReader, BytesIO
+from typing import TYPE_CHECKING, List, Optional, Tuple, Union
 
 import PIL.Image
 import PIL.ImageDraw
-import wand.image
+from wand.image import Color as WandColor  # type: ignore
+from wand.image import Image as WandImage
 
 from . import utils
-from .table import TableFinder
+from ._typing import T_bbox, T_num, T_obj, T_obj_list, T_point, T_seq
+from .table import T_table_settings, Table, TableFinder, TableSettings
+
+if TYPE_CHECKING:  # pragma: nocover
+    from pandas.core.frame import DataFrame
+    from pandas.core.series import Series
+
+    from .page import Page
 
 
-class COLORS(object):
+class COLORS:
     RED = (255, 0, 0)
     GREEN = (0, 255, 0)
     BLUE = (0, 0, 255)
@@ -20,35 +29,45 @@ DEFAULT_STROKE = COLORS.RED + (200,)
 DEFAULT_STROKE_WIDTH = 1
 DEFAULT_RESOLUTION = 72
 
+T_color = Union[Tuple[int, int, int], Tuple[int, int, int, int], str]
+T_contains_points = Union[Tuple[T_point, ...], List[T_point], T_obj]
 
-def get_page_image(stream, page_no, resolution):
+
+def get_page_image(stream: BufferedReader, page_no: int, resolution: int) -> WandImage:
     # If we are working with a file object saved to disk
     if hasattr(stream, "name"):
-        spec = dict(filename=f"{stream.name}[{page_no}]")
+        filename = f"{stream.name}[{page_no}]"
+        file = None
 
-        def postprocess(img):
+        def postprocess(img: WandImage) -> WandImage:
             return img
 
     # If we instead are working with a BytesIO stream
     else:
         stream.seek(0)
-        spec = dict(file=stream)
+        filename = None
+        file = stream
 
-        def postprocess(img):
-            return wand.image.Image(image=img.sequence[page_no])
+        def postprocess(img: WandImage) -> WandImage:
+            return WandImage(image=img.sequence[page_no])
 
-    with wand.image.Image(resolution=resolution, **spec) as img_init:
+    with WandImage(resolution=resolution, filename=filename, file=file) as img_init:
         img = postprocess(img_init)
         if img.alpha_channel:
-            img.background_color = wand.image.Color("white")
+            img.background_color = WandColor("white")
             img.alpha_channel = "remove"
         with img.convert("png") as png:
             im = PIL.Image.open(BytesIO(png.make_blob()))
             return im.convert("RGB")
 
 
-class PageImage(object):
-    def __init__(self, page, original=None, resolution=DEFAULT_RESOLUTION):
+class PageImage:
+    def __init__(
+        self,
+        page: "Page",
+        original: Optional[WandImage] = None,
+        resolution: int = DEFAULT_RESOLUTION,
+    ):
         self.page = page
         if original is None:
             self.original = get_page_image(
@@ -74,13 +93,13 @@ class PageImage(object):
             self.original = self.original.crop(tuple(map(int, cropbox)))
         self.reset()
 
-    def _reproject_bbox(self, bbox):
+    def _reproject_bbox(self, bbox: T_bbox) -> T_bbox:
         x0, top, x1, bottom = bbox
         _x0, _top = self._reproject((x0, top))
         _x1, _bottom = self._reproject((x1, bottom))
         return (_x0, _top, _x1, _bottom)
 
-    def _reproject(self, coord):
+    def _reproject(self, coord: T_point) -> T_point:
         """
         Given an (x0, top) tuple from the *root* coordinate system,
         return an (x0, top) tuple in the *image* coordinate system.
@@ -92,21 +111,25 @@ class PageImage(object):
         _top = (top + rtop - ptop) * self.scale
         return (_x0, _top)
 
-    def reset(self):
+    def reset(self) -> "PageImage":
         self.annotated = PIL.Image.new(self.original.mode, self.original.size)
         self.annotated.paste(self.original)
         self.draw = PIL.ImageDraw.Draw(self.annotated, "RGBA")
+        self.save = self.annotated.save
         return self
 
-    def copy(self):
+    def copy(self) -> "PageImage":
         return self.__class__(self.page, self.original)
 
     def draw_line(
-        self, points_or_obj, stroke=DEFAULT_STROKE, stroke_width=DEFAULT_STROKE_WIDTH
-    ):
+        self,
+        points_or_obj: T_contains_points,
+        stroke: T_color = DEFAULT_STROKE,
+        stroke_width: int = DEFAULT_STROKE_WIDTH,
+    ) -> "PageImage":
         if isinstance(points_or_obj, (tuple, list)):
             points = points_or_obj
-        elif type(points_or_obj) == dict and "points" in points_or_obj:
+        elif isinstance(points_or_obj, dict) and "points" in points_or_obj:
             points = points_or_obj["points"]
         else:
             obj = points_or_obj
@@ -116,42 +139,63 @@ class PageImage(object):
         )
         return self
 
-    def draw_lines(self, list_of_lines, **kwargs):
+    def draw_lines(
+        self,
+        list_of_lines: Union[T_seq[T_contains_points], "DataFrame"],
+        stroke: T_color = DEFAULT_STROKE,
+        stroke_width: int = DEFAULT_STROKE_WIDTH,
+    ) -> "PageImage":
         for x in utils.to_list(list_of_lines):
-            self.draw_line(x, **kwargs)
+            self.draw_line(x, stroke=stroke, stroke_width=stroke_width)
         return self
 
     def draw_vline(
-        self, location, stroke=DEFAULT_STROKE, stroke_width=DEFAULT_STROKE_WIDTH
-    ):
+        self,
+        location: T_num,
+        stroke: T_color = DEFAULT_STROKE,
+        stroke_width: int = DEFAULT_STROKE_WIDTH,
+    ) -> "PageImage":
         points = (location, self.page.bbox[1], location, self.page.bbox[3])
         self.draw.line(self._reproject_bbox(points), fill=stroke, width=stroke_width)
         return self
 
-    def draw_vlines(self, locations, **kwargs):
-        for x in utils.to_list(locations):
-            self.draw_vline(x, **kwargs)
+    def draw_vlines(
+        self,
+        locations: Union[List[T_num], "Series"],
+        stroke: T_color = DEFAULT_STROKE,
+        stroke_width: int = DEFAULT_STROKE_WIDTH,
+    ) -> "PageImage":
+        for x in list(locations):
+            self.draw_vline(x, stroke=stroke, stroke_width=stroke_width)
         return self
 
     def draw_hline(
-        self, location, stroke=DEFAULT_STROKE, stroke_width=DEFAULT_STROKE_WIDTH
-    ):
+        self,
+        location: T_num,
+        stroke: T_color = DEFAULT_STROKE,
+        stroke_width: int = DEFAULT_STROKE_WIDTH,
+    ) -> "PageImage":
         points = (self.page.bbox[0], location, self.page.bbox[2], location)
         self.draw.line(self._reproject_bbox(points), fill=stroke, width=stroke_width)
         return self
 
-    def draw_hlines(self, locations, **kwargs):
-        for x in utils.to_list(locations):
-            self.draw_hline(x, **kwargs)
+    def draw_hlines(
+        self,
+        locations: Union[List[T_num], "Series"],
+        stroke: T_color = DEFAULT_STROKE,
+        stroke_width: int = DEFAULT_STROKE_WIDTH,
+    ) -> "PageImage":
+        for x in list(locations):
+            self.draw_hline(x, stroke=stroke, stroke_width=stroke_width)
         return self
 
     def draw_rect(
         self,
-        bbox_or_obj,
-        fill=DEFAULT_FILL,
-        stroke=DEFAULT_STROKE,
-        stroke_width=DEFAULT_STROKE_WIDTH,
-    ):
+        bbox_or_obj: Union[T_bbox, T_obj],
+        fill: T_color = DEFAULT_FILL,
+        stroke: T_color = DEFAULT_STROKE,
+        stroke_width: int = DEFAULT_STROKE_WIDTH,
+    ) -> "PageImage":
         if isinstance(bbox_or_obj, (tuple, list)):
             bbox = bbox_or_obj
         else:
@@ -179,15 +223,25 @@ class PageImage(object):
             self.draw_lines(segments, stroke=stroke, stroke_width=stroke_width)
         return self
 
-    def draw_rects(self, list_of_rects, **kwargs):
+    def draw_rects(
+        self,
+        list_of_rects: Union[List[T_bbox], T_obj_list, "DataFrame"],
+        fill: T_color = DEFAULT_FILL,
+        stroke: T_color = DEFAULT_STROKE,
+        stroke_width: int = DEFAULT_STROKE_WIDTH,
+    ) -> "PageImage":
         for x in utils.to_list(list_of_rects):
-            self.draw_rect(x, **kwargs)
+            self.draw_rect(x, fill=fill, stroke=stroke, stroke_width=stroke_width)
         return self
 
     def draw_circle(
-        self, center_or_obj, radius=5, fill=DEFAULT_FILL, stroke=DEFAULT_STROKE
-    ):
-        if isinstance(center_or_obj, (tuple, list)):
+        self,
+        center_or_obj: Union[T_point, T_obj],
+        radius: int = 5,
+        fill: T_color = DEFAULT_FILL,
+        stroke: T_color = DEFAULT_STROKE,
+    ) -> "PageImage":
+        if isinstance(center_or_obj, tuple):
             center = center_or_obj
         else:
             obj = center_or_obj
@@ -197,17 +251,24 @@ class PageImage(object):
         self.draw.ellipse(self._reproject_bbox(bbox), fill, stroke)
         return self
 
-    def draw_circles(self, list_of_circles, **kwargs):
+    def draw_circles(
+        self,
+        list_of_circles: Union[List[T_point], T_obj_list, "DataFrame"],
+        radius: int = 5,
+        fill: T_color = DEFAULT_FILL,
+        stroke: T_color = DEFAULT_STROKE,
+    ) -> "PageImage":
         for x in utils.to_list(list_of_circles):
-            self.draw_circle(x, **kwargs)
+            self.draw_circle(x, radius=radius, fill=fill, stroke=stroke)
         return self
 
-    def save(self, *args, **kwargs):
-        return self.annotated.save(*args, **kwargs)
-
     def debug_table(
-        self, table, fill=DEFAULT_FILL, stroke=DEFAULT_STROKE, stroke_width=1
-    ):
+        self,
+        table: Table,
+        fill: T_color = DEFAULT_FILL,
+        stroke: T_color = DEFAULT_STROKE,
+        stroke_width: int = 1,
+    ) -> "PageImage":
         """
         Outline all found tables.
         """
@@ -216,24 +277,26 @@ class PageImage(object):
         )
         return self
 
-    def debug_tablefinder(self, tf={}):
+    def debug_tablefinder(
+        self, tf: Optional[Union[TableFinder, TableSettings, T_table_settings]] = None
+    ) -> "PageImage":
         if isinstance(tf, TableFinder):
-            pass
-        elif isinstance(tf, dict):
-            tf = self.page.debug_tablefinder(tf)
+            finder = tf
+        elif isinstance(tf, (TableSettings, dict)):
+            finder = self.page.debug_tablefinder(tf)
         else:
             raise ValueError(
                 "Argument must be instance of TableFinder"
                 "or a TableFinder settings dict."
             )
 
-        for table in tf.tables:
+        for table in finder.tables:
             self.debug_table(table)
 
-        self.draw_lines(tf.edges, stroke_width=1)
+        self.draw_lines(finder.edges, stroke_width=1)
 
         self.draw_circles(
-            tf.intersections.keys(),
+            list(finder.intersections.keys()),
             fill=COLORS.TRANSPARENT,
             stroke=COLORS.BLUE + (200,),
             radius=3,
@@ -242,12 +305,12 @@ class PageImage(object):
 
     def outline_words(
         self,
-        stroke=DEFAULT_STROKE,
-        fill=DEFAULT_FILL,
-        stroke_width=DEFAULT_STROKE_WIDTH,
-        x_tolerance=utils.DEFAULT_X_TOLERANCE,
-        y_tolerance=utils.DEFAULT_Y_TOLERANCE,
-    ):
+        stroke: T_color = DEFAULT_STROKE,
+        fill: T_color = DEFAULT_FILL,
+        stroke_width: int = DEFAULT_STROKE_WIDTH,
+        x_tolerance: T_num = utils.DEFAULT_X_TOLERANCE,
+        y_tolerance: T_num = utils.DEFAULT_Y_TOLERANCE,
+    ) -> "PageImage":
 
         words = self.page.extract_words(
             x_tolerance=x_tolerance, y_tolerance=y_tolerance
@@ -257,17 +320,17 @@ class PageImage(object):
 
     def outline_chars(
         self,
-        stroke=(255, 0, 0, 255),
-        fill=(255, 0, 0, int(255 / 4)),
-        stroke_width=DEFAULT_STROKE_WIDTH,
-    ):
+        stroke: T_color = (255, 0, 0, 255),
+        fill: T_color = (255, 0, 0, int(255 / 4)),
+        stroke_width: int = DEFAULT_STROKE_WIDTH,
+    ) -> "PageImage":
 
         self.draw_rects(
             self.page.chars, stroke=stroke, fill=fill, stroke_width=stroke_width
         )
         return self
 
-    def _repr_png_(self):
+    def _repr_png_(self) -> bytes:
         b = BytesIO()
         self.annotated.save(b, "PNG")
         return b.getvalue()
